# yellowbrick.features
# Visualizers for feature analysis and diagnostics.
#
# Author:   Benjamin Bengfort
# Created:  Mon Oct 03 21:30:18 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [0f4b236] benjamin@bengfort.com $

"""
Visualizers for feature analysis and diagnostics.
"""

##########################################################################
## Imports
##########################################################################

## Hoist visualizers into the features namespace
from .pcoords import ParallelCoordinates, parallel_coordinates
from .radviz import RadialVisualizer, RadViz, radviz
from .rankd import Rank1D, rank1d, Rank2D, rank2d
from .jointplot import JointPlot, JointPlotVisualizer, joint_plot
from .pca import PCA, PCADecomposition, pca_decomposition
from .manifold import Manifold, manifold_embedding

# Alias the TargetType defined in yellowbrick.utils.target
from yellowbrick.utils.target import TargetType

# RFECV and Feature Importances moved to model selection module as of YB v1.0
from yellowbrick.model_selection.rfecv import RFECV, rfecv
from yellowbrick.model_selection.importances import FeatureImportances
from yellowbrick.model_selection.importances import feature_importances



# yellowbrick.classifier.class_prediction_error
# Shows the balance of classes and their associated predictions.
#
# Author:   Larry Gray
# Author:   Benjamin Bengfort
# Created:  Fri Jul 20 10:26:25 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: class_prediction_error.py [] lwgray@gmail.com $

"""
Shows the balance of classes and their associated predictions.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from sklearn.utils.multiclass import unique_labels

from yellowbrick.draw import bar_stack
from yellowbrick.classifier.base import ClassificationScoreVisualizer
from yellowbrick.exceptions import ModelError, YellowbrickValueError, NotFitted

try:
    # See #1124: this allows compatibility for scikit-learn >= 0.20
    from sklearn.metrics._classification import _check_targets
except ImportError:
    from sklearn.metrics.classification import _check_targets


##########################################################################
## Class Prediction Error Chart
##########################################################################


class ClassPredictionError(ClassificationScoreVisualizer):
    """
    Class Prediction Error chart that shows the support for each class in the
    fitted classification model displayed as a stacked bar. Each bar is segmented
    to show the distribution of predicted classes for each class. It is initialized
    with a fitted model and generates a class prediction error chart on draw.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    classes : list of str, defult: None
        The class labels to use for the legend ordered by the index of the sorted
        classes discovered in the ``fit()`` method. Specifying classes in this
        manner is used to change the class names to a more specific format or
        to label encoded integer classes. Some visualizers may also use this
        field to filter the visualization for specific classes. For more advanced
        usage specify an encoder rather than class labels.

    encoder : dict or LabelEncoder, default: None
        A mapping of classes to human readable labels. Often there is a mismatch
        between desired class labels and those contained in the target variable
        passed to ``fit()`` or ``score()``. The encoder disambiguates this mismatch
        ensuring that classes are labeled correctly in the visualization.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The class labels observed while fitting.

    class_count_ : ndarray of shape (n_classes,)
        Number of samples encountered for each class during fitting.

    score_ : float
        An evaluation metric of the classifier on test data produced when
        ``score()`` is called. This metric is between 0 and 1 -- higher scores are
        generally better. For classifiers, this score is usually accuracy, but
        ensure you check the underlying model for more details about the score.

    predictions_ : ndarray
        An ndarray of predictions whose rows are the true classes and
        whose columns are the predicted classes
    """

    def __init__(
        self,
        estimator,
        ax=None,
        classes=None,
        encoder=None,
        is_fitted="auto",
        force_model=False,
        **kwargs
    ):
        super(ClassPredictionError, self).__init__(
            estimator,
            ax=ax,
            classes=classes,
            encoder=encoder,
            is_fitted=is_fitted,
            force_model=force_model,
            **kwargs
        )

    def score(self, X, y):
        """
        Generates a 2D array where each row is the count of the
        predicted classes and each column is the true class

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        Returns
        -------
        score_ : float
            Global accuracy score
        """
        # Must be computed before calling super
        # We're relying on predict to raise NotFitted
        y_pred = self.predict(X)
        y_type, y_true, y_pred = _check_targets(y, y_pred)
        if y_type not in ("binary", "multiclass"):
            raise YellowbrickValueError("{} is not supported".format(y_type))

        # Get the indices of the unique labels
        indices = unique_labels(y_true, y_pred)
        labels = self._labels()

        # Call super to compute self.score_ and verify classes
        try:
            super(ClassPredictionError, self).score(X, y)
        except ModelError as e:
            # raise visualizer-specific errors
            if labels is not None and len(labels) < len(indices):
                raise NotImplementedError(
                    "filtering classes is currently not supported"
                )
            else:
                raise e

        # Ensure all labels are used
        if labels is not None and len(labels) > len(indices):
            raise ModelError(
                "y and y_pred contain zero values for one of the specified classes"
            )

        # Create a table of predictions whose rows are the true classes
        # and whose columns are the predicted classes; each element
        # is the count of predictions for that class that match the true
        # value of that class.
        self.predictions_ = np.array(
            [
                [(y_pred[y == label_t] == label_p).sum() for label_p in indices]
                for label_t in indices
            ]
        )

        self.draw()
        return self.score_

    def draw(self):
        """
        Renders the class prediction error across the axis.

        Returns
        -------
        ax : Matplotlib Axes
            The axes on which the figure is plotted
        """

        if not hasattr(self, "predictions_") or not hasattr(self, "classes_"):
            raise NotFitted.from_estimator(self, "draw")

        legend_kws = {"bbox_to_anchor": (1.04, 0.5), "loc": "center left"}
        bar_stack(
            self.predictions_,
            self.ax,
            labels=list(self.classes_),
            ticks=self.classes_,
            colors=self.class_colors_,
            legend_kws=legend_kws,
        )
        return self.ax

    def finalize(self, **kwargs):
        """
        Adds a title and axis labels to the visualizer, ensuring that the
        y limits zoom the visualization in to the area of interest. Finalize
        also calls tight layout to ensure that no parts of the figure are
        cut off.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """

        # Set the title
        self.set_title("Class Prediction Error for {}".format(self.name))

        # Set the axes labels
        self.ax.set_xlabel("actual class")
        self.ax.set_ylabel("number of predicted class")

        # Compute the ceiling for the y limit
        cmax = max([sum(predictions) for predictions in self.predictions_])
        self.ax.set_ylim(0, cmax + cmax * 0.1)

        # Ensure the legend fits on the figure
        self.fig.tight_layout(rect=[0, 0, 0.90, 1])


##########################################################################
## Quick Method
##########################################################################


def class_prediction_error(
    estimator,
    X_train,
    y_train,
    X_test=None,
    y_test=None,
    ax=None,
    classes=None,
    encoder=None,
    is_fitted="auto",
    force_model=False,
    show=True,
    **kwargs
):
    """Class Prediction Error

    Divides the dataset X and y into train and test splits, fits the model on the train
    split, then scores the model on the test split. The visualizer displays the support
    for each class in the fitted classification model displayed as a stacked bar plot.
    Each bar is segmented to show the distribution of predicted classes for each class.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    X_train : ndarray or DataFrame of shape n x m
        A feature array of n instances with m features the model is trained on.
        Used to fit the visualizer and also to score the visualizer if test splits are
        not directly specified.

    y_train : ndarray or Series of length n
        An array or series of target or class values. Used to fit the visualizer and
        also to score the visualizer if test splits are not specified.

    X_test : ndarray or DataFrame of shape n x m, default: None
        An optional feature array of n instances with m features that the model
        is scored on if specified, using X_train as the training data.

    y_test : ndarray or Series of length n, default: None
        An optional array or series of target or class values that serve as actual
        labels for X_test for scoring purposes.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    classes : list of str, defult: None
        The class labels to use for the legend ordered by the index of the sorted
        classes discovered in the ``fit()`` method. Specifying classes in this
        manner is used to change the class names to a more specific format or
        to label encoded integer classes. Some visualizers may also use this
        field to filter the visualization for specific classes. For more advanced
        usage specify an encoder rather than class labels.

    encoder : dict or LabelEncoder, default: None
        A mapping of classes to human readable labels. Often there is a mismatch
        between desired class labels and those contained in the target variable
        passed to ``fit()`` or ``score()``. The encoder disambiguates this mismatch
        ensuring that classes are labeled correctly in the visualization.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs: dict
        Keyword arguments passed to the visualizer base classes.

    Returns
    -------
    viz : ClassPredictionError
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    viz = ClassPredictionError(
        estimator=estimator,
        ax=ax,
        classes=classes,
        encoder=encoder,
        is_fitted=is_fitted,
        force_model=force_model,
        **kwargs
    )

    # Fit the visualizer (calls draw)
    viz.fit(X_train, y_train, **kwargs)

    # Score the visualizer
    if X_test is not None and y_test is not None:
        viz.score(X_test, y_test)
    elif X_test is not None or y_test is not None:
        raise YellowbrickValueError("must specify both X_test and y_test or neither")
    else:
        viz.score(X_train, y_train)

    # Draw the final visualization
    if show:
        viz.show()
    else:
        viz.finalize()

    # Return the visualizer
    return viz



# yellowbrick.features.manifold
# Use manifold algorithms for high dimensional visualization.
#
# Author:  Benjamin Bengfort
# Created: Sat May 12 11:25:24 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: manifold.py [02f8c27] benjamin@bengfort.com $

"""
Use manifold algorithms for high dimensional visualization.
"""

##########################################################################
## Imports
##########################################################################

import warnings

from yellowbrick.utils.timer import Timer
from yellowbrick.utils.types import is_estimator
from yellowbrick.exceptions import ModelError, NotFitted
from yellowbrick.features.projection import ProjectionVisualizer
from yellowbrick.exceptions import YellowbrickValueError, YellowbrickWarning

from sklearn.base import clone
from sklearn.exceptions import NotFittedError
from sklearn.manifold import LocallyLinearEmbedding
from sklearn.manifold import Isomap, MDS, TSNE, SpectralEmbedding


##########################################################################
## Supported manifold algorithms by name lookup
##########################################################################

MANIFOLD_ALGORITHMS = {
    "lle": LocallyLinearEmbedding(method="standard", eigen_solver="auto"),
    "ltsa": LocallyLinearEmbedding(method="ltsa", eigen_solver="auto"),
    "hessian": LocallyLinearEmbedding(method="hessian", eigen_solver="auto"),
    "modified": LocallyLinearEmbedding(method="modified", eigen_solver="auto"),
    "isomap": Isomap(),
    "mds": MDS(),
    "spectral": SpectralEmbedding(),
    "tsne": TSNE(init="pca"),
}

MANIFOLD_NAMES = {
    "lle": "Locally Linear Embedding",
    "ltsa": "LTSA LLE",
    "hessian": "Hessian LLE",
    "modified": "Modified LLE",
    "isomap": "Isomap",
    "mds": "MDS",
    "spectral": "Spectral Embedding",
    "tsne": "t-SNE",
}


##########################################################################
## Manifold Embeddings
##########################################################################


class Manifold(ProjectionVisualizer):
    """
    The Manifold visualizer provides high dimensional visualization for feature
    analysis by embedding data into 2 dimensions using the sklearn.manifold
    package for manifold learning. In brief, manifold learning algorithms are
    unsuperivsed approaches to non-linear dimensionality reduction (unlike PCA
    or SVD) that help visualize latent structures in data.

    The manifold algorithm used to do the embedding in scatter plot space can
    either be a transformer or a string representing one of the already
    specified manifolds as follows:

        ============== ==========================
        Manifold       Description
        -------------- --------------------------
        ``"lle"``      `Locally Linear Embedding`_
        ``"ltsa"``     `LTSA LLE`_
        ``"hessian"``  `Hessian LLE`_
        ``"modified"`` `Modified LLE`_
        ``"isomap"``   `Isomap`_
        ``"mds"``      `Multi-Dimensional Scaling`_
        ``"spectral"`` `Spectral Embedding`_
        ``"tsne"``     `t-SNE`_
        ============== ==========================

    Each of these algorithms embeds non-linear relationships in different ways,
    allowing for an exploration of various structures in the feature space.
    Note however, that each of these algorithms has different time, memory and
    complexity requirements; take special care when using large datasets!

    The Manifold visualizer also shows the specified target (if given) as the
    color of the scatter plot. If a classification or clustering target is
    given, then discrete colors will be used with a legend. If a regression or
    continuous target is specified, then a colormap and colorbar will be shown.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None, the current axes will be used
        or generated if required.

    manifold : str or Transformer, default: "mds"
        Specify the manifold algorithm to perform the embedding. Either one of
        the strings listed in the table above, or an actual scikit-learn
        transformer. The constructed manifold is accessible with the manifold
        property, so as to modify hyperparameters before fit.

    n_neighbors : int, default: None
        Many manifold algorithms are nearest neighbors based, for those that
        are, this parameter specfies the number of neighbors to use in the
        embedding. If n_neighbors is not specified for those embeddings, it is
        set to 5 and a warning is issued. If the manifold algorithm doesn't use
        nearest neighbors, then this parameter is ignored.

    features : list, default: None
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    classes : list, default: None
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised. This parameter is only used in the
        discrete target type case and is ignored otherwise.

    colors : list or tuple, default: None
        A single color to plot all instances as or a list of colors to color each
        instance according to its class in the discrete case or as an ordered
        colormap in the sequential case. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        The colormap used to create the individual colors. In the discrete case
        it is used to compute the number of colors needed for each class and
        in the continuous case it is used to create a sequential color map based
        on the range of the target.

    target_type : str, default: "auto"
        Specify the type of target as either "discrete" (classes) or "continuous"
        (real numbers, usually for regression). If "auto", then it will
        attempt to determine the type by counting the number of unique values.

        If the target is discrete, the colors are returned as a dict with classes
        being the keys. If continuous the colors will be list having value of
        color for each point. In either case, if no target is specified, then
        color will be specified as the first color in the color cycle.

    projection : int or string, default: 2
        The number of axes to project into, either 2d or 3d. To plot 3d plots
        with matplotlib, please ensure a 3d axes is passed to the visualizer,
        otherwise one will be created using the current figure.

    alpha : float, default: 0.75
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    random_state : int or RandomState, default: None
        Fixes the random state for stochastic manifold algorithms.

    colorbar : bool, default: True
        If the target_type is "continous" draw a colorbar to the right of the
        scatter plot. The colobar axes is accessible using the cax property.

    kwargs : dict
        Keyword arguments passed to the base class and may influence the
        feature visualization properties.

    Attributes
    ----------
    fit_time_ : yellowbrick.utils.timer.Timer
        The amount of time in seconds it took to fit the Manifold.

    classes_ : ndarray, shape (n_classes,)
        The class labels that define the discrete values in the target. Only
        available if the target type is discrete. This is guaranteed to be
        strings even if the classes are a different type.

    features_ : ndarray, shape (n_features,)
        The names of the features discovered or used in the visualizer that
        can be used as an index to access or modify data in X. If a user passes
        feature names in, those features are used. Otherwise the columns of a
        DataFrame are used or just simply the indices of the data array.

    range_ : (min y, max y)
        A tuple that describes the minimum and maximum values in the target.
        Only available if the target type is continuous.

    Examples
    --------

    >>> viz = Manifold(manifold='isomap', target='discrete')
    >>> viz.fit_transform(X, y)
    >>> viz.show()

    Notes
    -----
    Specifying the target as ``'continuous'`` or ``'discrete'`` will influence
    how the visualizer is finally displayed, don't rely on the automatic
    determination from the Manifold!

    Scaling your data with the standard scalar before applying it to the
    visualizer is a great way of increasing performance. Additionally using
    the ``SelectKBest`` transformer may also improve performance and lead to
    better visualizations.

    .. warning::
        Manifold visualizers have extremly varying time, resource, and
        complexity requirements. Sampling data or features may be necessary
        in order to finish a manifold computation.

    .. seealso::
        The Scikit-Learn discussion on `Manifold Learning <http://scikit-learn.org/stable/modules/manifold.html>`_.

    .. _`Locally Linear Embedding`: http://scikit-learn.org/stable/modules/manifold.html#locally-linear-embedding
    .. _`LTSA LLE`: http://scikit-learn.org/stable/modules/manifold.html#local-tangent-space-alignment
    .. _`Hessian LLE`: http://scikit-learn.org/stable/modules/manifold.html#hessian-eigenmapping
    .. _`Modified LLE`: http://scikit-learn.org/stable/modules/manifold.html#modified-locally-linear-embedding
    .. _`Isomap`: http://scikit-learn.org/stable/modules/manifold.html#isomap
    .. _`Multi-Dimensional Scaling`: http://scikit-learn.org/stable/modules/manifold.html#multi-dimensional-scaling-mds
    .. _`Spectral Embedding`: http://scikit-learn.org/stable/modules/manifold.html#spectral-embedding
    .. _`t-SNE`: http://scikit-learn.org/stable/modules/manifold.html#t-distributed-stochastic-neighbor-embedding-t-sne
    """

    ALGORITHMS = MANIFOLD_ALGORITHMS

    def __init__(
        self,
        ax=None,
        manifold="mds",
        n_neighbors=None,
        features=None,
        classes=None,
        colors=None,
        colormap=None,
        target_type="auto",
        projection=2,
        alpha=0.75,
        random_state=None,
        colorbar=True,
        **kwargs
    ):

        super(Manifold, self).__init__(
            ax,
            features,
            classes,
            colors,
            colormap,
            target_type,
            projection,
            alpha,
            colorbar,
            **kwargs
        )
        self._name = None
        self._manifold = None

        self.n_neighbors = n_neighbors
        self.random_state = random_state
        self.manifold = manifold  # must be set last

    @property
    def manifold(self):
        """
        Property containing the manifold transformer constructed from the
        supplied hyperparameter. Use this property to modify the manifold
        before fit with ``manifold.set_params()``.
        """
        return self._manifold

    @manifold.setter
    def manifold(self, transformer):
        """
        Creates the manifold estimator if a string value is passed in,
        validates other objects passed in.
        """
        if not is_estimator(transformer):
            if transformer not in self.ALGORITHMS:
                raise YellowbrickValueError(
                    "could not create manifold for '{}'".format(str(transformer))
                )

            # 2 components is required for 2D plots
            n_components = self.projection
            requires_default_neighbors = {
                "lle",
                "ltsa",
                "isomap",
                "hessian",
                "spectral",
                "modified",
            }

            # Check if the n_neighbors attribute needs to be set.
            if self.n_neighbors is None and transformer in requires_default_neighbors:
                if transformer == "hessian":
                    self.n_neighbors = int(
                        1 + (n_components * (1 + (n_components + 1) / 2))
                    )
                else:
                    self.n_neighbors = 5

                # Issue a warning that the n_neighbors was set to a default.
                warnmsg = (
                    "using n_neighbors={};"
                    " please explicitly specify for the '{}' manifold"
                ).format(self.n_neighbors, str(transformer))
                warnings.warn(warnmsg, YellowbrickWarning)

            # Create a new transformer with the specified params
            self._name = MANIFOLD_NAMES[transformer]
            transformer = clone(self.ALGORITHMS[transformer])
            params = {
                "n_components": n_components,
                "n_neighbors": self.n_neighbors,
                "random_state": self.random_state,
            }

            for param in list(params.keys()):
                if param not in transformer.get_params():
                    del params[param]

            transformer.set_params(**params)

        self._manifold = transformer
        if self._name is None:
            self._name = self._manifold.__class__.__name__

    def fit(self, X, y=None, **kwargs):
        """
        Fits the manifold on X and transforms the data to plot it on the axes.
        See fit_transform() for more details.

        Parameters
        ----------
        X : array-like of shape (n, m)
            A matrix or data frame with n instances and m features

        y : array-like of shape (n,), optional
            A vector or series with target values for each instance in X. This
            vector is used to determine the color of the points in X.

        Returns
        -------
        self : Manifold
            Returns the visualizer object.

        """
        if not hasattr(self.manifold, "transform"):
            name = self.manifold.__class__.__name__
            raise ModelError(
                (
                    "{} requires data to be simultaneously fit and transformed, "
                    "use fit_transform instead"
                ).format(name)
            )

        # Call super to compute features, classes, colors, etc.
        super(Manifold, self).fit(X, y)
        with Timer() as self.fit_time_:
            self.manifold.fit(X)
        return self

    def fit_transform(self, X, y=None, **kwargs):
        """
        Fits the manifold on X and transforms the data to plot it on the axes.
        The optional y specified can be used to declare discrete colors. If
        the target is set to 'auto', this method also determines the target
        type, and therefore what colors will be used.

        Note also that fit records the amount of time it takes to fit the
        manifold and reports that information in the visualization.

        Parameters
        ----------
        X : array-like of shape (n, m)
            A matrix or data frame with n instances and m features

        y : array-like of shape (n,), optional
            A vector or series with target values for each instance in X. This
            vector is used to determine the color of the points in X.

        Returns
        -------
        Xprime : array-like of shape (n, 2)
            Returns the 2-dimensional embedding of the instances.

        """
        # Because some manifolds do not have transform, we cannot call individual
        # fit and transform methods, but must do it manually here.

        # Call super fit to compute features, classes, colors, etc.
        super(Manifold, self).fit(X, y)
        with Timer() as self.fit_time_:
            Xp = self.manifold.fit_transform(X)
        self.draw(Xp, y)
        return Xp

    def transform(self, X, y=None, **kwargs):
        """
        Returns the transformed data points from the manifold embedding.

        Parameters
        ----------
        X : array-like of shape (n, m)
            A matrix or data frame with n instances and m features

        y : array-like of shape (n,), optional
            The target, used to specify the colors of the points.

        Returns
        -------
        Xprime : array-like of shape (n, 2)
            Returns the 2-dimensional embedding of the instances.

        Note
        ----
        This method does not work with MDS, TSNE and SpectralEmbedding because
        it is yet to be implemented in sklearn.
        """
        # Because some manifolds do not have transform we cannot call super
        try:
            Xp = self.manifold.transform(X)
            self.draw(Xp, y)
            return Xp
        except NotFittedError:
            raise NotFitted.from_estimator(self, "transform")
        except AttributeError:
            name = self.manifold.__class__.__name__
            raise ModelError(
                (
                    "{} requires data to be simultaneously fit and transformed, "
                    "use fit_transform instead"
                ).format(name)
            )

        return Xp

    def draw(self, Xp, y=None):
        # Calls draw method from super class which draws scatter plot.
        super(Manifold, self).draw(Xp, y)
        return self.ax

    def finalize(self):
        """
        Add title and modify axes to make the image ready for display.
        """
        self.set_title(
            "{} Manifold (fit in {:0.2f} seconds)".format(
                self._name, self.fit_time_.interval
            )
        )
        self.ax.set_xlabel("Using {} features".format(len(self.features_)))
        # Draws legend for discrete target and colorbar for continuous.
        super(Manifold, self).finalize()


##########################################################################
## Quick Method
##########################################################################


def manifold_embedding(
    X,
    y=None,
    ax=None,
    manifold="mds",
    n_neighbors=None,
    features=None,
    classes=None,
    colors=None,
    colormap=None,
    target_type="auto",
    projection=2,
    alpha=0.75,
    random_state=None,
    colorbar=True,
    show=True,
    **kwargs
):
    """Quick method for Manifold visualizer.

    The Manifold visualizer provides high dimensional visualization for feature
    analysis by embedding data into 2 dimensions using the sklearn.manifold
    package for manifold learning. In brief, manifold learning algorithms are
    unsuperivsed approaches to non-linear dimensionality reduction (unlike PCA
    or SVD) that help visualize latent structures in data.

    .. seealso:: See Manifold for more details.

    Parameters
    ----------
    X : array-like of shape (n, m)
        A matrix or data frame with n instances and m features where m > 2.

    y : array-like of shape (n,), optional
        A vector or series with target values for each instance in X. This
        vector is used to determine the color of the points in X.

    ax : matplotlib.Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    manifold : str or Transformer, default: "lle"
        Specify the manifold algorithm to perform the embedding. Either one of
        the strings listed in the table above, or an actual scikit-learn
        transformer. The constructed manifold is accessible with the manifold
        property, so as to modify hyperparameters before fit.

    n_neighbors : int, default: None
        Many manifold algorithms are nearest neighbors based, for those that
        are, this parameter specfies the number of neighbors to use in the
        embedding. If n_neighbors is not specified for those embeddings, it is
        set to 5 and a warning is issued. If the manifold algorithm doesn't use
        nearest neighbors, then this parameter is ignored.

    features : list, default: None
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    classes : list, default: None
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised. This parameter is only used in the
        discrete target type case and is ignored otherwise.

    colors : list or tuple, default: None
        A single color to plot all instances as or a list of colors to color each
        instance according to its class in the discrete case or as an ordered
        colormap in the sequential case. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        The colormap used to create the individual colors. In the discrete case
        it is used to compute the number of colors needed for each class and
        in the continuous case it is used to create a sequential color map based
        on the range of the target.

    target_type : str, default: "auto"
        Specify the type of target as either "discrete" (classes) or "continuous"
        (real numbers, usually for regression). If "auto", then it will
        attempt to determine the type by counting the number of unique values.

        If the target is discrete, the colors are returned as a dict with classes
        being the keys. If continuous the colors will be list having value of
        color for each point. In either case, if no target is specified, then
        color will be specified as the first color in the color cycle.

    projection : int or string, default: 2
        The number of axes to project into, either 2d or 3d. To plot 3d plots
        with matplotlib, please ensure a 3d axes is passed to the visualizer,
        otherwise one will be created using the current figure.

    alpha : float, default: 0.75
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    random_state : int or RandomState, default: None
        Fixes the random state for stochastic manifold algorithms.

    colorbar : bool, default: True
        If the target_type is "continous" draw a colorbar to the right of the
        scatter plot. The colobar axes is accessible using the cax property.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments passed to the base class and may influence the
        feature visualization properties.

    Returns
    -------
    viz : Manifold
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    viz = Manifold(
        ax=ax,
        manifold=manifold,
        n_neighbors=n_neighbors,
        features=features,
        classes=classes,
        colors=colors,
        colormap=colormap,
        target_type=target_type,
        projection=projection,
        alpha=alpha,
        random_state=random_state,
        colorbar=colorbar,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    viz.fit_transform(X, y)

    if show:
        viz.show()
    else:
        viz.finalize()

    # Return the visualizer object
    return viz



# yellowbrick.download
# Downloads the example datasets for running the examples.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Created:  Wed May 18 11:54:45 2016 -0400
#
# Copyright (C) 2016 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: download.py [1f73d2b] benjamin@bengfort.com $

"""
Downloads the example datasets for running the examples.
"""

##########################################################################
## Imports
##########################################################################

import os
import argparse

from yellowbrick.datasets import get_data_home
from yellowbrick.datasets.loaders import DATASETS
from yellowbrick.datasets.download import download_data
from yellowbrick.datasets.path import cleanup_dataset


##########################################################################
## Functions
##########################################################################


def download_all(data_home=None, replace=False):
    """
    Downloads all the example datasets to the data directory specified by
    ``get_data_home``. This function ensures that all datasets are available
    for use with the examples.
    """
    for _, meta in DATASETS.items():
        download_data(
            meta["url"], meta["signature"], data_home=data_home, replace=replace
        )

    print(
        "Downloaded {} datasets to {}".format(len(DATASETS), get_data_home(data_home))
    )


def cleanup_all(data_home=None):
    """
    Cleans up all the example datasets in the data directory specified by
    ``get_data_home`` either to clear up disk space or start from fresh.
    """
    removed = 0
    for name, meta in DATASETS.items():
        _, ext = os.path.splitext(meta["url"])
        removed += cleanup_dataset(name, data_home=data_home, ext=ext)

    print(
        "Removed {} fixture objects from {}".format(removed, get_data_home(data_home))
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Yellowbrick data downloader",
        epilog="for troubleshooting please visit our GitHub issues",
    )
    parser.add_argument(
        "-c",
        "--cleanup",
        action="store_true",
        default=False,
        help="cleanup any existing datasets before download",
    )
    parser.add_argument(
        "--no-download",
        action="store_true",
        default=False,
        help="prevent new data from being downloaded",
    )
    parser.add_argument(
        "-f",
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite any existing data with new download",
    )
    parser.add_argument(
        "data_home",
        default=None,
        nargs="?",
        help="specify the data download location or set $YELLOWBRICK_DATA",
    )

    args = parser.parse_args()

    if args.cleanup:
        cleanup_all(data_home=args.data_home)

    if not args.no_download:
        download_all(data_home=args.data_home, replace=args.overwrite)



# yellowbrick.contrib.statsmodels
# Implements wrappers around hte statsmodels library to use Yellowbrick with.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Wed Apr 04 13:13:24 2018 -0400
#
# ID: __init__.py [d6ebc39] benjamin@bengfort.com $

"""
Implements wrappers around hte statsmodels library to use Yellowbrick with.
"""

##########################################################################
## Imports
##########################################################################

from .base import StatsModelsWrapper



# yellowbrick.utils.nan_warnings
# Small helpers that help find and filter missing data.
#
# Author:   Aylr
# Created:  Thu Dec 28 11:37:42 2017 -0700
#
# Copyright (C) 2018 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: nan_warnings.py [d2276d6] Aylr@users.noreply.github.com $
#
"""
Small helpers that help find and filter missing data.
"""

import numpy as np
import warnings
from yellowbrick.exceptions import DataWarning


def filter_missing(X, y=None):
    """
    Removes rows that contain np.nan values in data. If y is given,
    X and y will be filtered together so that their shape remains identical.
    For example, rows in X with nans will also remove rows in y, or rows in y
    with np.nans will also remove corresponding rows in X.

    Parameters
    ------------
    X : array-like
        Data in shape (m, n) that possibly contains np.nan values

    y : array-like, optional
        Data in shape (m, 1) that possibly contains np.nan values

    Returns
    --------
    X' : np.array
       Possibly transformed X with any row containing np.nan removed

    y' : np.array
        If y is given, will also return possibly transformed y to match the
        shape of X'.

    Notes
    ------
    This function will return either a np.array if only X is passed or a tuple
    if both X and y is passed. Because all return values are indexable, it is
    important to recognize what is being passed to the function to determine
    its output.
    """
    if y is not None:
        return filter_missing_X_and_y(X, y)
    else:
        return X[~np.isnan(X).any(axis=1)]


def filter_missing_X_and_y(X, y):
    """Remove rows from X and y where either contains nans."""
    y_nans = np.isnan(y)
    x_nans = np.isnan(X).any(axis=1)
    unioned_nans = np.logical_or(x_nans, y_nans)

    return X[~unioned_nans], y[~unioned_nans]


def warn_if_nans_exist(X):
    """Warn if nans exist in a numpy array."""
    null_count = count_rows_with_nans(X)
    total = len(X)
    percent = 100 * null_count / total

    if null_count > 0:
        warning_message = (
            "Warning! Found {} rows of {} ({:0.2f}%) with nan values. Only "
            "complete rows will be plotted.".format(null_count, total, percent)
        )
        warnings.warn(warning_message, DataWarning)


def count_rows_with_nans(X):
    """Count the number of rows in 2D arrays that contain any nan values."""
    if X.ndim == 2:
        return np.where(np.isnan(X).sum(axis=1) != 0, 1, 0).sum()


def count_nan_elements(data):
    """Count the number of elements in 1D arrays that are nan values."""
    if data.ndim == 1:
        return np.isnan(data).sum()



# yellowbrick.style.utils
# Utility functions for styles
#
# Author:   Neal Humphrey
# Created:  Wed Mar 22 12:39:35 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: utils.py [45268fc] humphrey.neal@gmail.com $

"""
Utility functions for styles
"""

##########################################################################
## Imports
##########################################################################

import numpy as np


def find_text_color(base_color, dark_color="black", light_color="white", coef_choice=0):
    """
    Takes a background color and returns the appropriate light or dark text color.
    Users can specify the dark and light text color, or accept the defaults of 'black' and 'white'

    base_color: The color of the background. This must be
        specified in RGBA with values between 0 and 1 (note, this is the default
        return value format of a call to base_color = cmap(number) to get the
        color corresponding to a desired number). Note, the value of `A` in RGBA
        is not considered in determining light/dark.

    dark_color: Any valid matplotlib color value.
        Function will return this value if the text should be colored dark

    light_color: Any valid matplotlib color value.
        Function will return this value if thet text should be colored light.

    coef_choice: slightly different approaches to calculating brightness. Currently two options in
        a list, user can enter 0 or 1 as list index. 0 is default.
    """

    # Coefficients:
    # option 0: http://www.nbdtech.com/Blog/archive/2008/04/27/Calculating-the-Perceived-Brightness-of-a-Color.aspx
    # option 1: http://stackoverflow.com/questions/596216/formula-to-determine-brightness-of-rgb-color
    coef_options = [
        np.array((0.241, 0.691, 0.068, 0)),
        np.array((0.299, 0.587, 0.114, 0)),
    ]

    coefs = coef_options[coef_choice]
    rgb = np.array(base_color) * 255
    brightness = np.sqrt(np.dot(coefs, rgb ** 2))

    # Threshold from option 0 link; determined by trial and error.
    # base is light
    if brightness > 130:
        return dark_color
    return light_color



# yellowbrick.datasets.path
# Helper functions for looking up dataset paths.
#
# Author:  Benjamin Bengfort
# Created: Thu Jul 26 14:10:51 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: path.py [7082742] benjamin@bengfort.com $

"""
Helper functions for looking up dataset paths.
"""

##########################################################################
## Imports
##########################################################################

import os
import shutil

from .signature import sha256sum
from yellowbrick.exceptions import DatasetsError


##########################################################################
## Fixtures
##########################################################################

FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")


##########################################################################
## Dataset path utilities
##########################################################################


def get_data_home(path=None):
    """
    Return the path of the Yellowbrick data directory. This folder is used by
    dataset loaders to avoid downloading data several times.

    By default, this folder is colocated with the code in the install directory
    so that data shipped with the package can be easily located. Alternatively
    it can be set by the ``$YELLOWBRICK_DATA`` environment variable, or
    programmatically by giving a folder path. Note that the ``'~'`` symbol is
    expanded to the user home directory, and environment variables are also
    expanded when resolving the path.
    """
    if path is None:
        path = os.environ.get("YELLOWBRICK_DATA", FIXTURES)

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    if not os.path.exists(path):
        os.makedirs(path)

    return path


def find_dataset_path(dataset, data_home=None, fname=None, ext=".csv.gz", raises=True):
    """
    Looks up the path to the dataset specified in the data home directory,
    which is found using the ``get_data_home`` function. By default data home
    is colocated with the code, but can be modified with the YELLOWBRICK_DATA
    environment variable, or passing in a different directory.

    The file returned will be by default, the name of the dataset in compressed
    CSV format. Other files and extensions can be passed in to locate other data
    types or auxilliary files.

    If the dataset is not found a ``DatasetsError`` is raised by default.

    Parameters
    ----------
    dataset : str
        The name of the dataset; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    fname : str, optional
        The filename to look up in the dataset path, by default it will be the
        name of the dataset. The fname must include an extension.

    ext : str, default: ".csv.gz"
        The extension of the data to look up in the dataset path, if the fname
        is specified then the ext parameter is ignored. If ext is None then
        the directory of the dataset will be returned.

    raises : bool, default: True
        If the path does not exist, raises a DatasetsError unless this flag is set
        to False, at which point None is returned (e.g. for checking if the
        path exists or not).

    Returns
    -------
    path : str or None
        A path to the requested file, guaranteed to exist if an exception is
        not raised during processing of the request (unless None is returned).

    raises : DatasetsError
        If raise is True and the path does not exist, raises a DatasetsError.
    """
    # Figure out the root directory of the datasets
    data_home = get_data_home(data_home)

    # Figure out the relative path to the dataset
    if fname is None:
        if ext is None:
            path = os.path.join(data_home, dataset)
        else:
            path = os.path.join(data_home, dataset, "{}{}".format(dataset, ext))
    else:
        path = os.path.join(data_home, dataset, fname)

    # Determine if the path exists
    if not os.path.exists(path):

        # Suppress exceptions if required
        if not raises:
            return None

        raise DatasetsError(
            ("could not find dataset at {} - does it need to be downloaded?").format(
                path
            )
        )

    return path


def dataset_exists(dataset, data_home=None):
    """
    Checks to see if a directory with the name of the specified dataset exists
    in the data home directory, found with ``get_data_home``.

    Parameters
    ----------
    dataset : str
        The name of the dataset; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    Returns
    -------
    exists : bool
        If a folder with the dataset name is in the data home directory.
    """
    data_home = get_data_home(data_home)
    path = os.path.join(data_home, dataset)

    return os.path.exists(path) and os.path.isdir(path)


def dataset_archive(dataset, signature, data_home=None, ext=".zip"):
    """
    Checks to see if the dataset archive file exists in the data home directory,
    found with ``get_data_home``. By specifying the signature, this function
    also checks to see if the archive is the latest version by comparing the
    sha256sum of the local archive with the specified signature.

    Parameters
    ----------
    dataset : str
        The name of the dataset; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable.

    signature : str
        The SHA 256 signature of the dataset, used to determine if the archive
        is the latest version of the dataset or not.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    ext : str, default: ".zip"
        The extension of the archive file.

    Returns
    -------
    exists : bool
        True if the dataset archive exists and is the latest version.
    """
    data_home = get_data_home(data_home)
    path = os.path.join(data_home, dataset + ext)

    if os.path.exists(path) and os.path.isfile(path):
        return sha256sum(path) == signature

    return False


def cleanup_dataset(dataset, data_home=None, ext=".zip"):
    """
    Removes the dataset directory and archive file from the data home directory.

    Parameters
    ----------
    dataset : str
        The name of the dataset; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    ext : str, default: ".zip"
        The extension of the archive file.

    Returns
    -------
    removed : int
        The number of objects removed from data_home.
    """
    removed = 0
    data_home = get_data_home(data_home)

    # Paths to remove
    datadir = os.path.join(data_home, dataset)
    archive = os.path.join(data_home, dataset + ext)

    # Remove directory and contents
    if os.path.exists(datadir):
        shutil.rmtree(datadir)
        removed += 1

    # Remove the archive file
    if os.path.exists(archive):
        os.remove(archive)
        removed += 1

    return removed



# yellowbrick.features.jointplot
# Implementation of joint plots for univariate and bivariate analysis.
#
# Author:   Prema Damodaran Roman
# Created:  Mon Apr 10 21:00:54 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers.
# For license information, see LICENSE.txt
#
# ID: jointplot.py [7f47800] pdamodaran@users.noreply.github.com $

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.pyplot as plt

try:
    # Only available in Matplotlib >= 2.0.2
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    make_axes_locatable = None

from .base import FeatureVisualizer

# from ..bestfit import draw_best_fit # TODO: return in #728
from ..utils.types import is_dataframe
from ..exceptions import YellowbrickValueError
from scipy.stats import pearsonr, spearmanr, kendalltau


# Default Colors
# TODO: should we reuse these colors?
FACECOLOR = "#FAFAFA"
HISTCOLOR = "#6897bb"


# Objects for export
__all__ = ["JointPlot", "JointPlotVisualizer", "joint_plot"]


##########################################################################
## Joint Plot Visualizer
##########################################################################


class JointPlot(FeatureVisualizer):
    """
    Joint plots are useful for machine learning on multi-dimensional data, allowing for
    the visualization of complex interactions between different data dimensions, their
    varying distributions, and even their relationships to the target variable for
    prediction.

    The Yellowbrick ``JointPlot`` can be used both for pairwise feature analysis and
    feature-to-target plots. For pairwise feature analysis, the ``columns`` argument can
    be used to specify the index of the two desired columns in ``X``. If ``y`` is also
    specified, the plot can be colored with a heatmap or by class. For feature-to-target
    plots, the user can provide either ``X`` and ``y`` as 1D vectors, or a ``columns``
    argument with an index to a single feature in ``X`` to be plotted against ``y``.

    Histograms can be included by setting the ``hist`` argument to ``True`` for a
    frequency distribution, or to ``"density"`` for a probability density function. Note
    that histograms requires matplotlib 2.0.2 or greater.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes will be
        used (or generated if required). This is considered the base axes where the
        the primary joint plot is drawn. It will be shifted and two additional axes
        added above (xhax) and to the right (yhax) if hist=True.

    columns : int, str, [int, int], [str, str], default: None
        Determines what data is plotted in the joint plot and acts as a selection index
        into the data passed to ``fit(X, y)``. This data therefore must be indexable by
        the column type (e.g. an int for a numpy array or a string for a DataFrame).

        If None is specified then either both X and y must be 1D vectors and they will
        be plotted against each other or X must be a 2D array with only 2 columns. If a
        single index is specified then the data is indexed as ``X[columns]`` and plotted
        jointly with the target variable, y. If two indices are specified then they are
        both selected from X, additionally in this case, if y is specified, then it is
        used to plot the color of points.

        Note that these names are also used as the x and y axes labels if they aren't
        specified in the joint_kws argument.

    correlation : str, default: 'pearson'
        The algorithm used to compute the relationship between the variables in the
        joint plot, one of: 'pearson', 'covariance', 'spearman', 'kendalltau'.

    kind : str in {'scatter', 'hex'}, default: 'scatter'
        The type of plot to render in the joint axes. Note that when kind='hex' the
        target cannot be plotted by color.

    hist : {True, False, None, 'density', 'frequency'}, default: True
        Draw histograms showing the distribution of the variables plotted jointly.
        If set to 'density', the probability density function will be plotted.
        If set to True or 'frequency' then the frequency will be plotted.
        Requires Matplotlib >= 2.0.2.

    alpha : float, default: 0.65
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    {joint, hist}_kws : dict, default: None
        Additional keyword arguments for the plot components.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    corr_ : float
        The correlation or relationship of the data in the joint plot, specified by the
        correlation algorithm.

    Examples
    --------

    >>> viz = JointPlot(columns=["temp", "humidity"])
    >>> viz.fit(X, y)
    >>> viz.show()
    """

    # TODO: should we couple more closely with Rank2D?
    correlation_methods = {
        "pearson": lambda x, y: pearsonr(x, y)[0],
        "spearman": lambda x, y: spearmanr(x, y)[0],
        "covariance": lambda x, y: np.cov(x, y)[0, 1],
        "kendalltau": lambda x, y: kendalltau(x, y)[0],
    }

    def __init__(
        self,
        ax=None,
        columns=None,
        correlation="pearson",
        kind="scatter",
        hist=True,
        alpha=0.65,
        joint_kws=None,
        hist_kws=None,
        **kwargs
    ):
        # Initialize the visualizer
        super(JointPlot, self).__init__(ax=ax, **kwargs)
        self._xhax, self._yhax = None, None

        # Set and validate the columns
        self.columns = columns
        if self.columns is not None and not isinstance(self.columns, (int, str)):
            self.columns = tuple(self.columns)
            if len(self.columns) > 2:
                raise YellowbrickValueError(
                    (
                        "'{}' contains too many indices or is invalid for joint plot - "
                        "specify either a single int or str index or two columns as a list"
                    ).format(columns)
                )

        # Seet and validate the correlation
        self.correlation = correlation
        if self.correlation not in self.correlation_methods:
            raise YellowbrickValueError(
                "'{}' is an invalid correlation method, use one of {}".format(
                    self.correlation, ", ".join(self.correlation_methods.keys())
                )
            )

        # Set and validate the kind of plot
        self.kind = kind
        if self.kind not in {"scatter", "hex", "hexbin"}:
            raise YellowbrickValueError(
                ("'{}' is invalid joint plot kind, use 'scatter' or 'hex'").format(
                    self.kind
                )
            )

        # Set and validate the histogram if specified
        self.hist = hist
        if self.hist not in {True, "density", "frequency", None, False}:
            raise YellowbrickValueError(
                (
                    "'{}' is an invalid argument for hist, use None, True, "
                    "False, 'density', or 'frequency'"
                ).format(hist)
            )

        # If hist is True, test the version availability
        if self.hist in {True, "density", "frequency"}:
            self._layout()

        # Set the additional visual parameters
        self.alpha = alpha
        self.joint_kws = joint_kws
        self.hist_kws = hist_kws

    @property
    def xhax(self):
        """
        The axes of the histogram for the top of the JointPlot (X-axis)
        """
        if self._xhax is None:
            raise AttributeError(
                "this visualizer does not have a histogram for the X axis"
            )
        return self._xhax

    @property
    def yhax(self):
        """
        The axes of the histogram for the right of the JointPlot (Y-axis)
        """
        if self._yhax is None:
            raise AttributeError(
                "this visualizer does not have a histogram for the Y axis"
            )
        return self._yhax

    def _layout(self):
        """
        Creates the grid layout for the joint plot, adding new axes for the histograms
        if necessary and modifying the aspect ratio. Does not modify the axes or the
        layout if self.hist is False or None.
        """
        # Ensure the axes are created if not hist, then return.
        if not self.hist:
            self.ax
            return

        # Ensure matplotlib version compatibility
        if make_axes_locatable is None:
            raise YellowbrickValueError(
                (
                    "joint plot histograms requires matplotlib 2.0.2 or greater "
                    "please upgrade matplotlib or set hist=False on the visualizer"
                )
            )

        # Create the new axes for the histograms
        divider = make_axes_locatable(self.ax)
        self._xhax = divider.append_axes("top", size=1, pad=0.1, sharex=self.ax)
        self._yhax = divider.append_axes("right", size=1, pad=0.1, sharey=self.ax)

        # Modify the display of the axes
        self._xhax.xaxis.tick_top()
        self._yhax.yaxis.tick_right()
        self._xhax.grid(False, axis="y")
        self._yhax.grid(False, axis="x")

    def fit(self, X, y=None):
        """
        Fits the JointPlot, creating a correlative visualization between the columns
        specified during initialization and the data and target passed into fit:

            - If self.columns is None then X and y must both be specified as 1D arrays
              or X must be a 2D array with only 2 columns.
            - If self.columns is a single int or str, that column is selected to be
              visualized against the target y.
            - If self.columns is two ints or strs, those columns are visualized against
              each other. If y is specified then it is used to color the points.

        This is the main entry point into the joint plot visualization.

        Parameters
        ----------
        X : array-like
            An array-like object of either 1 or 2 dimensions depending on self.columns.
            Usually this is a 2D table with shape (n, m)

        y : array-like, default: None
            An vector or 1D array that has the same length as X. May be used to either
            directly plot data or to color data points.
        """
        # Convert python objects to numpy arrays
        if isinstance(X, (list, tuple)):
            X = np.array(X)

        if y is not None and isinstance(y, (list, tuple)):
            y = np.array(y)

        # Case where no columns are specified
        if self.columns is None:
            if (y is None and (X.ndim != 2 or X.shape[1] != 2)) or (
                y is not None and (X.ndim != 1 or y.ndim != 1)
            ):
                raise YellowbrickValueError(
                    (
                        "when self.columns is None specify either X and y as 1D arrays "
                        "or X as a matrix with 2 columns"
                    )
                )

            if y is None:
                # Draw the first column as x and the second column as y
                self.draw(X[:, 0], X[:, 1], xlabel="0", ylabel="1")
                return self

            # Draw x against y
            self.draw(X, y, xlabel="x", ylabel="y")
            return self

        # Case where a single string or int index is specified
        if isinstance(self.columns, (int, str)):
            if y is None:
                raise YellowbrickValueError(
                    "when self.columns is a single index, y must be specified"
                )

            # fetch the index from X -- raising index error if not possible
            x = self._index_into(self.columns, X)
            self.draw(x, y, xlabel=str(self.columns), ylabel="target")
            return self

        # Case where there is a double index for both columns
        columns = tuple(self.columns)
        if len(columns) != 2:
            raise YellowbrickValueError(
                ("'{}' contains too many indices or is invalid for joint plot").format(
                    columns
                )
            )

        # TODO: color the points based on the target if it is given
        x = self._index_into(columns[0], X)
        y = self._index_into(columns[1], X)
        self.draw(x, y, xlabel=str(columns[0]), ylabel=str(columns[1]))
        return self

    def draw(self, x, y, xlabel=None, ylabel=None):
        """
        Draw the joint plot for the data in x and y.

        Parameters
        ----------
        x, y : 1D array-like
            The data to plot for the x axis and the y axis

        xlabel, ylabel : str
            The labels for the x and y axes.
        """
        # This is a little weird to be here, but it is the best place to perform
        # this computation given how fit calls draw and returns.
        self.corr_ = self.correlation_methods[self.correlation](x, y)

        # First draw the joint plot
        joint_kws = self.joint_kws or {}
        joint_kws.setdefault("alpha", self.alpha)
        joint_kws.setdefault("label", "{}={:0.3f}".format(self.correlation, self.corr_))

        # Draw scatter joint plot
        if self.kind == "scatter":
            self.ax.scatter(x, y, **joint_kws)

            # TODO: Draw best fit line (or should this be kind='reg'?)

        # Draw hexbin joint plot
        elif self.kind in ("hex", "hexbin"):
            joint_kws.setdefault("mincnt", 1)
            joint_kws.setdefault("gridsize", 50)
            joint_kws.setdefault("cmap", "Blues")
            self.ax.hexbin(x, y, **joint_kws)

        # Something bad happened
        else:
            raise ValueError("unknown joint plot kind '{}'".format(self.kind))

        # Set the X and Y axis labels on the plot
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)

        # If we're not going to draw histograms, stop here
        if not self.hist:
            # Ensure the current axes is always the main joint plot axes
            plt.sca(self.ax)
            return self.ax

        # Draw the histograms
        hist_kws = self.hist_kws or {}
        hist_kws.setdefault("bins", 50)
        if self.hist == "density":
            hist_kws.setdefault("density", True)

        self.xhax.hist(x, **hist_kws)
        self.yhax.hist(y, orientation="horizontal", **hist_kws)

        # Ensure the current axes is always the main joint plot axes
        plt.sca(self.ax)
        return self.ax

    def finalize(self, **kwargs):
        """
        Finalize executes any remaining image modifications making it ready to show.
        """
        # Set the aspect ratio to make the visualization square
        # TODO: still unable to make plot square using make_axes_locatable
        # x0,x1 = self.ax.get_xlim()
        # y0,y1 = self.ax.get_ylim()
        # self.ax.set_aspect(abs(x1-x0)/abs(y1-y0))

        # Add the title to the plot if the user has set one.
        self.set_title("")

        # TODO: use manual legend so legend works with both scatter and hexbin
        # Set the legend with full opacity patches using manual legend.
        # Or Add the colorbar if this is a continuous plot.
        if self.kind == "scatter":
            self.ax.legend(loc="best", frameon=True)

        # Finalize the histograms
        if self.hist:
            plt.setp(self.xhax.get_xticklabels(), visible=False)
            plt.setp(self.yhax.get_yticklabels(), visible=False)
            plt.sca(self.ax)

        # Call tight layout to maximize readability
        self.fig.tight_layout()

    def _index_into(self, idx, data):
        """
        Attempts to get the column from the data using the specified index, raises an
        exception if this is not possible from this point in the stack.
        """
        try:
            if is_dataframe(data):
                # Assume column indexing
                return data[idx]
            # Otherwise assume numpy array-like indexing
            return data[:, idx]
        except Exception as e:
            raise IndexError(
                "could not index column '{}' into type {}: {}".format(
                    self.columns, data.__class__.__name__, e
                )
            )


# Alias for JointPlot
JointPlotVisualizer = JointPlot


##########################################################################
## Quick Method for JointPlot visualizations
##########################################################################


def joint_plot(
        X,
        y,
        ax=None,
        columns=None,
        correlation="pearson",
        kind="scatter",
        hist=True,
        alpha=0.65,
        joint_kws=None,
        hist_kws=None,
        show=True,
        **kwargs
):
    """
    Joint plots are useful for machine learning on multi-dimensional data, allowing for
    the visualization of complex interactions between different data dimensions, their
    varying distributions, and even their relationships to the target variable for
    prediction.

    The Yellowbrick ``JointPlot`` can be used both for pairwise feature analysis and
    feature-to-target plots. For pairwise feature analysis, the ``columns`` argument can
    be used to specify the index of the two desired columns in ``X``. If ``y`` is also
    specified, the plot can be colored with a heatmap or by class. For feature-to-target
    plots, the user can provide either ``X`` and ``y`` as 1D vectors, or a ``columns``
    argument with an index to a single feature in ``X`` to be plotted against ``y``.

    Histograms can be included by setting the ``hist`` argument to ``True`` for a
    frequency distribution, or to ``"density"`` for a probability density function. Note
    that histograms requires matplotlib 2.0.2 or greater.

    Parameters
    ----------
    X : array-like
        An array-like object of either 1 or 2 dimensions depending on self.columns.
        Usually this is a 2D table with shape (n, m)

    y : array-like, default: None
        An vector or 1D array that has the same length as X. May be used to either
        directly plot data or to color data points.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes will be
        used (or generated if required). This is considered the base axes where the
        the primary joint plot is drawn. It will be shifted and two additional axes
        added above (xhax) and to the right (yhax) if hist=True.

    columns : int, str, [int, int], [str, str], default: None
        Determines what data is plotted in the joint plot and acts as a selection index
        into the data passed to ``fit(X, y)``. This data therefore must be indexable by
        the column type (e.g. an int for a numpy array or a string for a DataFrame).

        If None is specified then either both X and y must be 1D vectors and they will
        be plotted against each other or X must be a 2D array with only 2 columns. If a
        single index is specified then the data is indexed as ``X[columns]`` and plotted
        jointly with the target variable, y. If two indices are specified then they are
        both selected from X, additionally in this case, if y is specified, then it is
        used to plot the color of points.

        Note that these names are also used as the x and y axes labels if they aren't
        specified in the joint_kws argument.

    correlation : str, default: 'pearson'
        The algorithm used to compute the relationship between the variables in the
        joint plot, one of: 'pearson', 'covariance', 'spearman', 'kendalltau'.

    kind : str in {'scatter', 'hex'}, default: 'scatter'
        The type of plot to render in the joint axes. Note that when kind='hex' the
        target cannot be plotted by color.

    hist : {True, False, None, 'density', 'frequency'}, default: True
        Draw histograms showing the distribution of the variables plotted jointly.
        If set to 'density', the probability density function will be plotted.
        If set to True or 'frequency' then the frequency will be plotted.
        Requires Matplotlib >= 2.0.2.

    alpha : float, default: 0.65
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    {joint, hist}_kws : dict, default: None
        Additional keyword arguments for the plot components.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    corr_ : float
        The correlation or relationship of the data in the joint plot, specified by the
        correlation algorithm.
    """
    visualizer = JointPlot(
        ax=ax,
        columns=columns,
        correlation=correlation,
        kind=kind,
        hist=hist,
        alpha=alpha,
        joint_kws=joint_kws,
        hist_kws=hist_kws,
        **kwargs
    )

    # Fit and show the visualizer
    visualizer.fit(X, y)
    if show:
        visualizer.show()
    else:
        visualizer.finalize()
    return visualizer



# yellowbrick.features.decomposition
#
# Author:   George Richardson
# Created:  Fri Mar 2 16:16:00 2018 +0000
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: decomposition.py [0ed6e8a] g.raymond.richardson@gmail.com $

##########################################################################
## Imports
##########################################################################

from yellowbrick.style import palettes
from yellowbrick.features.base import FeatureVisualizer

from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

##########################################################################
## Quick Methods
##########################################################################


def explained_variance_visualizer(
    X,
    y=None,
    ax=None,
    scale=True,
    center=True,
    colormap=palettes.DEFAULT_SEQUENCE,
    **kwargs
):
    """Produce a plot of the explained variance produced by a dimensionality
        reduction algorithm using n=1 to n=n_components dimensions. This is a single
        plot to help identify the best trade off between number of dimensions
        and amount of information retained within the data.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n rows with m features

        y : ndarray or Series of length n
            An array or Series of target or class values

        ax : matplotlib Axes, default: None
            The aces to plot the figure on

        scale : bool, default: True
            Boolean that indicates if the values of X should be scaled.

        colormap : string or cmap, default: None
            optional string or matplotlib cmap to colorize lines
            Use either color to colorize the lines on a per class basis or
            colormap to color them on a continuous scale.

        kwargs : dict
            Keyword arguments that are passed to the base class and may influence
            the visualization as defined in other Visualizers.

        Returns
        -------
        viz : ExplainedVariance
            Returns the fitted, finalized visualizer

        Examples
        --------
        >>> from sklearn import datasets
        >>> bc = datasets.load_breast_cancer()
        >>> X = bc = bc.data
        >>> explained_variance_visualizer(X, scale=True, center=True, colormap='RdBu_r')

        """

    # Instantiate the visualizer
    visualizer = ExplainedVariance(X=X)

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)
    visualizer.transform(X)
    visualizer.finalize()

    # Return the visualizer object
    return visualizer


##########################################################################
## Explained Variance Feature Visualizer
##########################################################################


class ExplainedVariance(FeatureVisualizer):
    """

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The aces to plot the figure on

    scale : bool, default: True
        Boolean that indicates if the values of X should be scaled.

    colormap : string or cmap, default: None
        optional string or matplotlib cmap to colorize lines
        Use either color to colorize the lines on a per class basis or
        colormap to color them on a continuous scale.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.


    Examples
    --------

    >>> visualizer = ExplainedVariance()
    >>> visualizer.fit(X)
    >>> visualizer.transform(X)
    >>> visualizer.show()

    """

    def __init__(
        self,
        ax=None,
        scale=True,
        center=True,
        n_components=None,
        colormap=palettes.DEFAULT_SEQUENCE,
        **kwargs
    ):

        super(ExplainedVariance, self).__init__(ax=ax, **kwargs)

        self.colormap = colormap
        self.n_components = n_components
        self.center = center
        self.scale = scale
        self.pipeline = Pipeline(
            [
                ("scale", StandardScaler(with_mean=self.center, with_std=self.scale)),
                ("pca", PCA(n_components=self.n_components)),
            ]
        )
        self.pca_features = None

    @property
    def explained_variance_(self):
        return self.pipeline.steps[-1][1].explained_variance_

    def fit(self, X, y=None):
        self.pipeline.fit(X)
        self.draw()
        return self

    def transform(self, X):
        self.pca_features = self.pipeline.transform(X)
        return self.pca_features

    def draw(self):
        X = self.explained_variance_
        self.ax.plot(X)
        return self.ax

    def finalize(self, **kwargs):
        # Set the title
        self.set_title("Explained Variance Plot")

        # Set the axes labels
        self.ax.set_ylabel("Explained Variance")
        self.ax.set_xlabel("Number of Components")



# yellowbrick.datasets.download
# Downloads the example datasets for running the examples.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Raul Peralta
# Created:  Wed May 18 11:54:45 2016 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: download.py [1f73d2b] benjamin@bengfort.com $

"""
Downloads the example datasets for running the examples.
"""

##########################################################################
## Imports
##########################################################################

import os
import zipfile

from urllib.request import urlopen

from .signature import sha256sum
from .path import get_data_home, cleanup_dataset

from yellowbrick.exceptions import DatasetsError


# Downlod chunk size
CHUNK = 524288


##########################################################################
## Download functions
##########################################################################


def download_data(url, signature, data_home=None, replace=False, extract=True):
    """
    Downloads the zipped data set specified at the given URL, saving it to
    the data directory specified by ``get_data_home``. This function verifies
    the download with the given signature and extracts the archive.

    Parameters
    ----------
    url : str
        The URL of the dataset on the Internet to GET

    signature : str
        The SHA 256 hash of the dataset archive being downloaded to verify
        that the dataset has been correctly downloaded

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    replace : bool, default: False
        If the data archive already exists, replace the dataset. If this is
        False and the dataset exists, an exception is raised.

    extract : bool, default: True
        Extract the archive file after downloading it
    """
    data_home = get_data_home(data_home)

    # Get the name of the file from the URL
    basename = os.path.basename(url)
    name, _ = os.path.splitext(basename)

    # Get the archive and data directory paths
    archive = os.path.join(data_home, basename)
    datadir = os.path.join(data_home, name)

    # If the archive exists cleanup or raise override exception
    if os.path.exists(archive):
        if not replace:
            raise DatasetsError(
                ("dataset already exists at {}, set replace=False to overwrite").format(
                    archive
                )
            )

        cleanup_dataset(name, data_home=data_home)

    # Create the output directory if it does not exist
    if not os.path.exists(datadir):
        os.mkdir(datadir)

    # Fetch the response in a streaming fashion and write it to disk.
    response = urlopen(url)

    with open(archive, "wb") as f:
        while True:
            chunk = response.read(CHUNK)
            if not chunk:
                break
            f.write(chunk)

    # Compare the signature of the archive to the expected one
    if sha256sum(archive) != signature:
        raise ValueError("Download signature does not match hardcoded signature!")

    # If extract, extract the zipfile.
    if extract:
        zf = zipfile.ZipFile(archive)
        zf.extractall(path=data_home)



# yellowbrick.contrib.prepredict
# PrePredict estimator allows Yellowbrick to work with results produced by an estimator.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Mon Jul 12 07:07:33 2021 -0400
#
# ID: prepredict.py [] benjamin@bengfort.com $

"""
PrePredict estimator allows Yellowbrick to work with results produced by an estimator
prior to the visual diagnostic workflow, particularly for inferences that require
extensive time or compute resources.
"""

##########################################################################
## Imports
##########################################################################

import pathlib
import numpy as np

from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score, r2_score, silhouette_score
from yellowbrick.contrib.wrapper import CLASSIFIER, CLUSTERER, REGRESSOR


class PrePredict(BaseEstimator):
    """
    The Passthrough estimator allows users to specify pre-predicted results to
    Yellowbrick without the need to input the original estimator. Note that Yellowbrick
    often uses the learned attributes of the estimator to produce rich visual
    diagnostics, so this estimator may not work for all Yellowbrick visualizers.

    The passthrough estimator can accept data either in memory as a numpy array or it
    can accept a string, which it interprets as a path on disk to load the data from.

    Currently passthrough does not support predict_proba or decision_function methods,
    which it could if it was passed predicted data as 2D array instead of a 1D array.

    Parameters
    ----------
    data : array-like, func, or file-like object, string, or pathlib.Path
        The predicted values wrapped by the estimator and returned on predict() and
        used by the score function. The default expectation is that data is a 1D numpy
        array of y_hat or y_pred values produced by some other estimator. Data can also
        be a func, which is called and returned, or a file-like object, string, or
        pathlib.Path at which point the data is loaded from disk using ``np.load``.

    estimator_type : str, optional
        One of "classifier", "regressor", "clusterer", "DensityEstimator", or
        "outlier_detector" that allows the contrib estimator to pass the scikit-learn
        ``is_classifier``, etc. functions. If not specified, the Yellowbrick visualizer
        you're trying to use may error.
    """

    def __init__(self, data, estimator_type=None):
        self.data = data
        self._estimator_type = estimator_type

    def fit(self, X, y=None):
        """
        Fit is a no-op, simply returning self per the scikit-learn API.
        """
        return self

    def predict(self, X):
        """
        Predict returns the embedded data but does not perform any checks on the
        validity of X (e.g. that it has the same shape as the internal data).
        """
        return self._load()

    def score(self, X, y=None):
        """
        Score uses an appropriate metric for the estimator type and compares the input
        y values with the pre-predicted values.
        """
        if self._estimator_type == CLASSIFIER:
            return accuracy_score(y, self._load())

        if self._estimator_type == REGRESSOR:
            return r2_score(y, self._load())

        if self._estimator_type == CLUSTERER:
            labels = y if y is not None else self._load()
            return silhouette_score(X, labels)

        # If the estimator type is unknown return NaN since the score can't be computed.
        return np.nan

    def _load(self):
        """
        Loads the data by performing type checking to determine if data is a callable
        whose result needs to be returned, or an argument that supports from disk
        loading. If neither of these things, then assumes the data is array-like and
        returns it directly.
        """
        if callable(self.data):
            return self.data()

        if hasattr(self.data, "read") or isinstance(self.data, (str, pathlib.Path)):
            return np.load(self.data)

        return self.data


# yellowbrick.contrib.scatter
# Implements a 2d scatter plot for feature analysis.
#
# Author:   Nathan Danielsen
# Created:  Fri Feb 26 19:40:00 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: scatter.py [a89633e] benjamin@bengfort.com $
"""
Implements a 2D scatter plot for feature analysis.
"""

##########################################################################
# Imports
##########################################################################

import itertools
import numpy as np

from yellowbrick.features.base import DataVisualizer
from yellowbrick.utils import is_dataframe, is_structured_array
from yellowbrick.utils import has_ndarray_int_columns
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.style.colors import resolve_colors


##########################################################################
# Quick Methods
##########################################################################


def scatterviz(
    X,
    y=None,
    ax=None,
    features=None,
    classes=None,
    color=None,
    colormap=None,
    markers=None,
    alpha=1.0,
    **kwargs
):
    """Displays a bivariate scatter plot.

    This helper function is a quick wrapper to utilize the ScatterVisualizer
    (Transformer) for one-off analysis.

    Parameters
    ----------

    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n, default: None
        An array or series of target or class values

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    features : list of strings, default: None
        The names of two features or columns.
        More than that will raise an error.

    classes : list of strings, default: None
        The names of the classes in the target

    color : list or tuple of colors, default: None
        Specify the colors for each individual class

    colormap : string or matplotlib cmap, default: None
        Sequential colormap for continuous target

    markers : iterable of strings, default: ,+o*vhd
        Matplotlib style markers for points on the scatter plot points

    alpha : float, default: 1.0
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    Returns
    -------
    viz : ScatterVisualizer
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    visualizer = ScatterVisualizer(
        ax=ax,
        features=features,
        classes=classes,
        color=color,
        colormap=colormap,
        markers=markers,
        alpha=alpha,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)
    visualizer.transform(X)

    # Return the visualizer object
    return visualizer


##########################################################################
# Static ScatterVisualizer Visualizer
##########################################################################


class ScatterVisualizer(DataVisualizer):
    """
    ScatterVisualizer is a bivariate feature data visualization algorithm that
    plots using the Cartesian coordinates of each point.

        Parameters
        ----------

        ax : a matplotlib plot, default: None
            The axis to plot the figure on.

        x : string, default: None
            The feature name that corresponds to a column name or index postion
            in the matrix that will be plotted against the x-axis

        y : string, default: None
            The feature name that corresponds to a column name or index postion
            in the matrix that will be plotted against the y-axis

        features : a list of two feature names to use, default: None
            List of two features that correspond to the columns in the array.
            The order of the two features correspond to X and Y axes on the
            graph. More than two feature names or columns will raise an error.
            If a DataFrame is passed to fit and features is None, feature names
            are selected that are the columns of the DataFrame.

        classes : a list of class names for the legend, default: None
            If classes is None and a y value is passed to fit then the classes
            are selected from the target vector.

        color : optional list or tuple of colors to colorize points, default: None
            Use either color to colorize the points on a per class basis or
            colormap to color them on a continuous scale.

        colormap : optional string or matplotlib cmap to colorize points, default: None
            Use either color to colorize the points on a per class basis or
            colormap to color them on a continuous scale.

        markers : iterable of strings, default: ,+o*vhd
            Matplotlib style markers for points on the scatter plot points

        alpha : float, default: 1.0
            Specify a transparency where 1 is completely opaque and 0 is completely
            transparent. This property makes densely clustered points more visible.

        kwargs : keyword arguments passed to the super class.

        These parameters can be influenced later on in the visualization
        process, but can and should be set as early as possible.
    """

    def __init__(
        self,
        ax=None,
        x=None,
        y=None,
        features=None,
        classes=None,
        color=None,
        colormap=None,
        markers=None,
        alpha=1.0,
        **kwargs
    ):
        """
        Initialize the base scatter with many of the options required in order
        to make the visualization work.
        """
        super(ScatterVisualizer, self).__init__(
            ax=ax,
            features=features,
            classes=classes,
            color=color,
            colormap=colormap,
            **kwargs
        )

        self.x = x
        self.y = y

        self.alpha = alpha

        self.markers = itertools.cycle(
            kwargs.pop("markers", (",", "+", "o", "*", "v", "h", "d"))
        )

        self.color = color
        self.colormap = colormap

        if self.x is not None and self.y is not None and self.features is not None:
            raise YellowbrickValueError("Please specify x,y or features, not both.")

        if self.x is not None and self.y is not None and self.features is None:
            self.features = [self.x, self.y]

        # Ensure with init that features doesn't have more than two features
        if features is not None:
            if len(features) != 2:
                raise YellowbrickValueError(
                    "ScatterVisualizer only accepts two features."
                )

    def fit(self, X, y=None, **kwargs):
        """
        The fit method is the primary drawing input for the parallel coords
        visualization since it has both the X and y data required for the
        viz and the transform method does not.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with 2 features

        y : ndarray or Series of length n
            An array or series of target or class values

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer
        """
        _, ncols = X.shape

        # NOTE: Do not call super for this class, it conflicts with the fit.
        # Setting these variables is similar to the old behavior of DataVisualizer.
        # TODO: refactor to make use of the new DataVisualizer functionality
        self.features_ = self.features
        self.classes_ = self.classes

        if ncols == 2:
            X_two_cols = X
            if self.features_ is None:
                self.features_ = ["Feature One", "Feature Two"]

        # Handle the feature names if they're None.
        elif self.features_ is not None and is_dataframe(X):
            X_two_cols = X[self.features_].values

        # handle numpy named/ structured array
        elif self.features_ is not None and is_structured_array(X):
            X_selected = X[self.features_]
            X_two_cols = X_selected.copy().view(
                (np.float64, len(X_selected.dtype.names))
            )

        # handle features that are numeric columns in ndarray matrix
        elif self.features_ is not None and has_ndarray_int_columns(self.features_, X):
            f_one, f_two = self.features_
            X_two_cols = X[:, [int(f_one), int(f_two)]]

        else:
            raise YellowbrickValueError(
                """
                ScatterVisualizer only accepts two features, please
                explicitly set these two features in the init kwargs or
                pass a matrix/ dataframe in with only two columns."""
            )

        # Store the classes for the legend if they're None.
        if self.classes_ is None:
            # TODO: Is this the most efficient method?
            self.classes_ = [str(label) for label in np.unique(y)]

        # Draw the instances
        self.draw(X_two_cols, y, **kwargs)

        # Fit always returns self.
        return self

    def draw(self, X, y, **kwargs):
        """Called from the fit method, this method creates a scatter plot that
        draws each instance as a class or target colored point, whose location
        is determined by the feature data set.
        """
        # Set the axes limits
        self.ax.set_xlim([-1, 1])
        self.ax.set_ylim([-1, 1])

        # set the colors
        color_values = resolve_colors(
            n_colors=len(self.classes_), colormap=self.colormap, colors=self.color
        )

        colors = dict(zip(self.classes_, color_values))

        # Create a data structure to hold the scatter plot representations
        to_plot = {}
        for kls in self.classes_:
            to_plot[kls] = [[], []]

        # Add each row of the data set to to_plot for plotting
        # TODO: make this an independent function for override
        for i, row in enumerate(X):
            row_ = np.repeat(np.expand_dims(row, axis=1), 2, axis=1)
            x_, y_ = row_[0], row_[1]
            kls = self.classes_[y[i]]

            to_plot[kls][0].append(x_)
            to_plot[kls][1].append(y_)

        # Add the scatter plots from the to_plot function
        # TODO: store these plots to add more instances to later
        # TODO: make this a separate function
        for i, kls in enumerate(self.classes_):
            self.ax.scatter(
                to_plot[kls][0],
                to_plot[kls][1],
                marker=next(self.markers),
                color=colors[kls],
                label=str(kls),
                alpha=self.alpha,
                **kwargs
            )

        self.ax.axis("equal")

    def finalize(self, **kwargs):
        """
        Adds a title and a legend and ensures that the axis labels are set as
        the feature names being visualized.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Divide out the two features
        feature_one, feature_two = self.features_

        # Set the title
        self.set_title(
            "Scatter Plot: {0} vs {1}".format(str(feature_one), str(feature_two))
        )
        # Add the legend
        self.ax.legend(loc="best")
        self.ax.set_xlabel(str(feature_one))
        self.ax.set_ylabel(str(feature_two))


# Alias for ScatterViz
ScatterViz = ScatterVisualizer
